/*
 * mod_xml_dtc provides an XML-DTC implementation as an apache module
 * Copyright (C) 2003-2004 Mike Hommey <mh@glandium.org>
 * 
 * This module is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This module is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this module; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 */

/* 
 *  To activate the module in Apache's apache.conf file, for instance
 *  for the URL /xml_dtc, add the following to your apache.conf:
 *
 *    #   apache.conf
 *    LoadModule xml_dtc_module libexec/mod_xml_dtc.so
 *    <Location /xml-dtc>
 *    XmlDtcEngine On
 *    XmlDtcSitemap /path/to/sitemap.xmap
 *    </Location>
 */ 

#include "httpd.h"
#include "http_config.h"
#include "http_protocol.h"
#include "http_log.h"
#include "ap_config.h"

#include <xml-dtc.h>
#include <config.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

module MODULE_VAR_EXPORT xml_dtc_module;

typedef struct {
  xml_dtc_sitemap *sitemap;
  time_t mtime;
} mod_xml_dtc_dir_config_data;

typedef struct {
  char *filename;
  int enabled;
  int root_hack;
  mod_xml_dtc_dir_config_data *data;
} mod_xml_dtc_dir_config;

#if 0
static int set_mime_type(void *context, char *mime_type, char *encoding) {
  ((request_rec *)context)->content_type = mime_type;
  return 0;
}

static int ap_rwrite_wrapper(void *context, char *buf, int len) {
  if (!((request_rec *)context)->header_only)
    return ap_rwrite(buf, len, (request_rec *)context);
  return 0;
}
#endif

static void apache_log_callback(server_rec *s, int level, char *format, va_list va) {
  char buf[1024];
  strncpy(buf, "mod_xml_dtc: ", 13);
  vsnprintf(buf + 13, 1010, format, va);
  ap_log_error(APLOG_MARK, level, s, buf);
}

static void xml_dtc_init_mod(server_rec *s) {
  if (!xml_dtc.resource) {
    xml_dtc_resource *resource;
    
    resource = xml_dtc_resource_new();
    xml_dtc_resource_set_msg_callback(resource, (xml_dtc_msg_callback)apache_log_callback);
    xml_dtc_resource_set_msg_context(resource, s);
    xml_dtc_resource_set_msg_channel(resource, XML_DTC_MSG_DEBUG, (void *)APLOG_DEBUG);
    xml_dtc_resource_set_msg_channel(resource, XML_DTC_MSG_INFO, (void *)APLOG_INFO);
    xml_dtc_resource_set_msg_channel(resource, XML_DTC_MSG_NOTICE, (void *)APLOG_NOTICE);
    xml_dtc_resource_set_msg_channel(resource, XML_DTC_MSG_WARNING, (void *)APLOG_WARNING);
    xml_dtc_resource_set_msg_channel(resource, XML_DTC_MSG_ERROR, (void *)APLOG_ERR);
    xml_dtc_resource_set_msg_verbosity(resource, XML_DTC_MSG_DEBUG);
    
    xml_dtc_init(resource);
    xml_dtc_init_modules();
  }
}

static int mod_xml_dtc_load_sitemap(mod_xml_dtc_dir_config *conf) {
  struct stat fstat;

  if (!conf->filename)
    return -1;

  if (stat(conf->filename, &fstat)) {
    xml_dtc_sitemap_free(conf->data->sitemap);
    conf->data->sitemap = NULL;
    return -1;
  }

  if ((conf->data->sitemap == NULL) || (fstat.st_mtime != conf->data->mtime)) {
    xml_dtc_msg(xml_dtc.resource, XML_DTC_MSG_NOTICE, "Loading sitemap %s in pid %d", conf->filename, getpid());
    xml_dtc_sitemap_free(conf->data->sitemap);
    conf->data->sitemap = xml_dtc_sitemap_new_from_file(xml_dtc.resource, conf->filename);
    conf->data->mtime = fstat.st_mtime;
  }
  if (conf->data->sitemap)
    return 0;
  else
    return -1;
}

static int register_headers(xml_dtc_request *request) {
  array_header *env = ap_table_elts(((request_rec *)request->context)->headers_in);
  table_entry *element = (table_entry *) env->elts;
  int i;

  for (i = 0; i < env->nelts; i++) {
    xml_dtc_request_set_header(request, element[i].key, element[i].val);
  }
  return 0;
}

int register_parameters(xml_dtc_request *request) {
  char *query_string = ((request_rec *)request->context)->args;

  return xml_dtc_request_set_parameters_from_query_string(request, query_string);
}

/* The sample content handler */
static int mod_xml_dtc_handler(request_rec *r)
{
  xml_dtc_request *request;
  mod_xml_dtc_dir_config *c;
  char tmp[255];
  int i;

  c = (mod_xml_dtc_dir_config *) ap_get_module_config(r->per_dir_config, &xml_dtc_module);

  if ((c == NULL) || (! c->enabled))
    return DECLINED;

  xml_dtc_init_mod(r->server);

  if (mod_xml_dtc_load_sitemap(c)) {
    return DECLINED;
  }

  if (c->root_hack)
    request = xml_dtc_request_new(xml_dtc.resource, r->method, r->uri, r->uri);
  else
    request = xml_dtc_request_new(xml_dtc.resource, r->method, r->uri, r->path_info);
  request->register_headers = register_headers;
  request->register_parameters = register_parameters;
  request->context = r;
  /*
  request->set_mime_type = set_mime_type;
  request->set_content_length = (int (*)(void *, int)) ap_set_content_length;
  request->write_context = r;
  request->write = ap_rwrite_wrapper;
  request->send_http_header = (int (*)(void *)) ap_send_http_header;
  */
  xml_dtc_sitemap_request(c->data->sitemap, request);

  if (request->status == 302) {
    r->status = 302;
  } else {
    if (request->error.type != NULL) {
      r->status = HTTP_NOT_FOUND;
    }

    if (request->result != NULL) {
      
      if (request->mime_type) {
	r->content_type = request->mime_type;
      } else {
	r->content_type = "text/plain";
      }
      if (request->encoding) {
	sprintf(tmp, "%s; charset=%s", r->content_type, request->encoding);
	r->content_type = tmp;
      }
      if (request->content_length >= 0)
	ap_set_content_length(r, request->content_length);
    }
  }
  if (request->out_headers)
    for(i = 0; i < request->out_headers->num; i++) {
      ap_table_setn(r->headers_out, request->out_headers->list[i].key, (char *) request->out_headers->list[i].element);
    }
  if (request->newcookies)
    for (i = 0; i < request->newcookies->num; i++) {
      /* BEURK */
      char buf[4096];
      snprintf(buf, 4096, "%s=%s", request->newcookies->list[i].key, (char *)request->newcookies->list[i].element);
      ap_table_setn(r->headers_out, "Set-Cookie", buf);
      
    }
  ap_send_http_header(r);
  if (!r->header_only)
    if (request->status != 302) {
      if (request->content_length >= 0)
	ap_rwrite(request->result, request->content_length, r);
      else
	ap_rprintf(r, "%s", request->result);
    }

  xml_dtc_request_free(request);

  return OK;
}

static const char *mod_xml_dtc_sitemap(cmd_parms *cmd, mod_xml_dtc_dir_config *conf, char *param) {
  /* Naughty way of doing things, since memory will never be released */
  xml_dtc_init_mod(cmd->server);
  conf->filename = param;
  if (cmd->path && (strncmp(cmd->path, "/", 2) != 0))
    conf->root_hack = 0;
  else
    conf->root_hack = 1;

  mod_xml_dtc_load_sitemap(conf);

  return NULL;
}

static const char *mod_xml_dtc_engine(cmd_parms *cmd, mod_xml_dtc_dir_config *conf, int flag) {
  conf->enabled = flag;

  return NULL;
}

static void *mod_xml_dtc_create_dir_conf(pool *p, char *d) {
  mod_xml_dtc_dir_config *c = (mod_xml_dtc_dir_config *) ap_pcalloc(p, sizeof(mod_xml_dtc_dir_config));
  memset(c, 0, sizeof(mod_xml_dtc_dir_config));
  c->data = (mod_xml_dtc_dir_config_data *) ap_pcalloc(p, sizeof(mod_xml_dtc_dir_config_data));
  return c;
}

static void *mod_xml_dtc_merge_dir_conf(pool *p, void *basev, void* overridesv) {
  mod_xml_dtc_dir_config *overrides = (mod_xml_dtc_dir_config *) overridesv;
  mod_xml_dtc_dir_config *a = (mod_xml_dtc_dir_config *) ap_pcalloc(p, sizeof(mod_xml_dtc_dir_config));

  a->enabled = overrides->enabled;
  a->filename = overrides->filename;
  a->data = overrides->data;
  a->root_hack = overrides->root_hack;
  return a;
}

static void *mod_xml_dtc_create_server_conf(pool *p, server_rec *s) {
  mod_xml_dtc_dir_config *c = (mod_xml_dtc_dir_config *) ap_pcalloc(p, sizeof(mod_xml_dtc_dir_config));
  memset(c, 0, sizeof(mod_xml_dtc_dir_config));
  c->data = (mod_xml_dtc_dir_config_data *) ap_pcalloc(p, sizeof(mod_xml_dtc_dir_config_data));
  return c;
}

static void *mod_xml_dtc_merge_server_conf(pool *p, void *basev, void* overridesv) {
  mod_xml_dtc_dir_config *overrides = (mod_xml_dtc_dir_config *) overridesv;
  mod_xml_dtc_dir_config *a = (mod_xml_dtc_dir_config *) ap_pcalloc(p, sizeof(mod_xml_dtc_dir_config));

  a->enabled = overrides->enabled;
  a->filename = overrides->filename;
  a->data = overrides->data;
  a->root_hack = overrides->root_hack;
  return a;
}

static const command_rec mod_xml_dtc_commands[] = {
  { "XmlDtcEngine", mod_xml_dtc_engine, NULL, OR_FILEINFO, FLAG,
    "On or Off to enable or disable (default) the xml-dtc engine" },
  { "XmlDtcSitemap", mod_xml_dtc_sitemap, NULL, OR_FILEINFO, TAKE1,
    "the path of the sitemap to use for the xml-dtc engine" },
  { NULL }
};

/* Dispatch list of content handlers */
static const handler_rec mod_xml_dtc_handlers[] = { 
    { "xml-dtc", mod_xml_dtc_handler }, 
    { NULL, NULL }
};

void mod_xml_dtc_init(server_rec *s, pool *p) {
  ap_add_version_component( PACKAGE_NAME "/" PACKAGE_VERSION );
}

static int mod_xml_dtc_trans(request_rec *r) {
  mod_xml_dtc_dir_config *c;
  c = (mod_xml_dtc_dir_config *) ap_get_module_config(r->per_dir_config, &xml_dtc_module);
  if (c)
    if (c->enabled) {
      r->handler = "xml-dtc";
      return OK;
    }

  return DECLINED;
}

/* Dispatch list for API hooks */
module MODULE_VAR_EXPORT xml_dtc_module = {
    STANDARD_MODULE_STUFF, 
    mod_xml_dtc_init,      /* module initializer                  */
    mod_xml_dtc_create_dir_conf,  /* create per-dir    config structures */
    mod_xml_dtc_merge_dir_conf,  /* merge  per-dir    config structures */
    mod_xml_dtc_create_server_conf,  /* create per-server config structures */
    mod_xml_dtc_merge_server_conf,  /* merge  per-server config structures */
    mod_xml_dtc_commands,  /* table of config file commands       */
    mod_xml_dtc_handlers,  /* [#8] MIME-typed-dispatched handlers */
    NULL,     /* [#1] URI to filename translation    */
    NULL,                  /* [#4] validate user id from request  */
    NULL,                  /* [#5] check if the user is ok _here_ */
    NULL,                  /* [#3] check access by host address   */
    NULL,                  /* [#6] determine MIME type            */
    mod_xml_dtc_trans,                  /* [#7] pre-run fixups                 */
    NULL,                  /* [#9] log a transaction              */
    NULL,                  /* [#2] header parser                  */
    NULL,                  /* child_init                          */
    NULL,                  /* child_exit                          */
    NULL                   /* [#0] post read-request              */
#ifdef EAPI
   ,NULL,                  /* EAPI: add_module                    */
    NULL,                  /* EAPI: remove_module                 */
    NULL,                  /* EAPI: rewrite_command               */
    NULL                   /* EAPI: new_connection                */
#endif
};

