/*
 * This module provides serialization to the XML-DTC framework
 * Copyright (C) 2003-2004 Mike Hommey <mh@glandium.org>
 * 
 * This module is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This module is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this module; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 */

#include <modules.h>
#include <xml-dtc.h>

typedef struct _xml_dtc_serializer_caller xml_dtc_serializer_caller;

typedef struct _xml_dtc_serializer_data xml_dtc_serializer_data;
struct _xml_dtc_serializer_data {
  xmlChar *doctype_public, *doctype_system, *encoding, *mime_type;
};

static int _xml_dtc_serializer_free(xml_dtc_serializer_data *data) {
  if (data) {
    if (data->doctype_public)
      xmlFree(data->doctype_public);
    if (data->doctype_system)
      xmlFree(data->doctype_system);
    if (data->encoding)
      xmlFree(data->encoding);
    if (data->mime_type)
      xmlFree(data->mime_type);

    free(data);
  }
  return 0;
}

static xml_dtc_component *xml_dtc_handler_serializer_callback(xml_dtc_handler *handler, xml_dtc_hash *handlers, xml_dtc_sitemap *sitemap, xmlNodePtr node) {
  xmlChar *attr;
  xml_dtc_component *result;
  xml_dtc_serializer_data *data;

  attr = xmlGetProp(node, "name");
  result = xml_dtc_component_new(sitemap->resource, attr, (char *)node->name);
  data = (xml_dtc_serializer_data *) malloc(sizeof(xml_dtc_serializer_data));
  memset(data, 0, sizeof(xml_dtc_serializer_data));
  result->_private = data;
  result->free_private = (int (*) (void*)) _xml_dtc_serializer_free;
  result->handle = handler->_private;
  free(attr);
  data->mime_type = xmlGetProp(node, "mime-type");

  node = node->children;

  while (node) {
    /* Ignore Blank Nodes */
    if (! xmlIsBlankNode(node)) {
      /* Test if node is an element or something else (PI, Comment...) */
      if (node->type != XML_ELEMENT_NODE) {
        /* Comments are allowed */
        if (node->type == XML_COMMENT_NODE) {
          xml_dtc_msg(sitemap->resource, XML_DTC_MSG_DEBUG, "Found comment");
        } else {
          xml_dtc_msg(sitemap->resource, XML_DTC_MSG_DEBUG, "Element or comment expected");
          return NULL;
        }
      } else {
        /* There must be NO namespace */
        if (node->ns) {
          xml_dtc_msg(sitemap->resource, XML_DTC_MSG_ERROR, "expecting no namespace, but got %s",
                      node->ns->href);
          return NULL;
        }
        if (xmlStrcmp(node->name, "doctype-public") == 0) {
          data->doctype_public = xmlNodeGetContent(node);
        } else if (xmlStrcmp(node->name, "doctype-system") == 0) {
          data->doctype_system = xmlNodeGetContent(node);
        } else if (xmlStrcmp(node->name, "encoding") == 0) {
          data->encoding = xmlNodeGetContent(node);
        } else {
          xml_dtc_msg(sitemap->resource, XML_DTC_MSG_ERROR, "Unauthorized configuration %s", node->name);
          return NULL;
        }
      }
    }
    node = node->next;
  }

  xml_dtc_sitemap_add(sitemap, result);
  return (xml_dtc_component *)handler;
}

static xml_dtc_component *xml_dtc_handler_serialize_callback(xml_dtc_handler *handler, xml_dtc_hash *handlers, xml_dtc_sitemap *sitemap, xmlNodePtr node) {
  xml_dtc_caller *result;

  result = xml_dtc_caller_new(sitemap->resource, "serializer", "serializer");
  result->component.name = xmlGetProp(node, "type");

  return (xml_dtc_component *)result;
}

static int xml_dtc_serialize(xml_dtc_caller *caller, xml_dtc_request *request) {
  if (! request->internal) {
    int size;
    xmlNodePtr node;
    xml_dtc_component *serializer = (xml_dtc_component *) caller->component.object;
    xml_dtc_serializer_data *data = serializer->_private;

    node = xmlDocGetRootElement(request->xmltree);

    if (data->doctype_public) {
      xmlCreateIntSubset(request->xmltree, node->name, data->doctype_public, data->doctype_system);
    }
    if (data->encoding)
      xmlDocDumpMemoryEnc(request->xmltree, &request->result, &size, data->encoding);
    else
      xmlDocDumpMemory(request->xmltree, &request->result, &size);
    xmlFreeDoc(request->xmltree);
    request->xmltree = NULL;

    request->content_length = size;
    request->mime_type = data->mime_type;
    request->encoding = data->encoding;
  }
  return 0;
}

static int xml_dtc_serializers_module_register(void) {
  return 0;
}

static const xml_dtc_caller_def xml_dtc_serializers_callers[] = {
  { "serialize", "serializer", (xml_dtc_handler_callback) xml_dtc_handler_serialize_callback },
  { NULL }
};

static const xml_dtc_component_def xml_dtc_serializers[] = {
  { "xml", "serializer", xml_dtc_serialize, (xml_dtc_handler_callback) xml_dtc_handler_serializer_callback },
  { NULL }
};

const xml_dtc_module xml_dtc_serializers_module = {
  XML_DTC_MODULE_STANDARD_STUFF,
  xml_dtc_serializers_module_register,
  NULL,
  xml_dtc_serializers_callers,
  xml_dtc_serializers
};
